package coffeetronix;

import java.io.File;
import java.io.FileWriter;
import java.io.FilenameFilter;
import java.io.IOException;
import java.util.Collection;
import java.util.Iterator;
import java.util.LinkedList;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;

import coffeetronix.diagram.*;

/** 
 * Holds a list of java source directories of a coffeetronix project 
 * parses all java files on opening and creates the java units, manages all 
 * diagram hints (positions of the  shapes).
 * 
 * @author Rene Stoeckel (rene.stoeckel-at-web.de)
 * 
 */
public class JavaProject
{
	DiagramPanel panel;

	private File file;

	private String projecthome= "";
	private LinkedList srcdirs= new LinkedList();
	private LinkedList widgethints= new LinkedList();
	private LinkedList linkhints= new LinkedList();
	private LinkedList javaunits= new LinkedList();
	private LinkedList comments= new LinkedList();

	/** holds the positions of the elements of the diagram */

	public JavaProject(DiagramPanel pan)
	{
		panel= pan;
	}
	public void init()
	{

		projecthome= "";

		srcdirs= new LinkedList();
		widgethints= new LinkedList();
		linkhints= new LinkedList();
		javaunits= new LinkedList();
		comments= new LinkedList();

	}

	public class ShapeHint
	{
		public String name= "";
		public int x, y;
		public int color= 0;
		public boolean isexpanded= true;

		public ShapeHint(String name, String x, String y, String isexpanded, String color)
		{
			this.name= name;
			try
			{
				this.x= Integer.parseInt(x);
				this.y= Integer.parseInt(y);
				this.isexpanded= !isexpanded.equals("false");
				this.color= Integer.parseInt(color);
			}
			catch (Exception e)
			{
			};
		}
	}

	/** Holds information about the lines in the diagram (type, from, to). */
	public class LinkHint
	{
		public LinkHint(String from, String to, String type)
		{
			this.from= from;
			this.to= to;
			this.type= type;
		}
		public String toString()
		{
			return "from=\"" + from + "\" to=\"" + to + "\" type=\"" + type + "\"";
		}
		String from, to, type;
	}

	private class Comment
	{
		public Comment(String acomment, String aname, String ato, String ax, String ay, int alines)
		{
			comment= acomment;
			name= aname;
			to= ato;
			try
			{
				this.x= Integer.parseInt(ax);
				this.y= Integer.parseInt(ay);
			}
			catch (Exception e)
			{
			};
			lines= alines;
		}

		String comment, name, to;
		int x, y, lines;
	}

	/** Constructs information for classdiagrampanel, classlist, linklist etc. */
	public void constructDiagram()
	{

		panel.displayOff();
		panel.clearAll();

		for (Iterator iter= javaunits.iterator(); iter.hasNext();)
		{
			try
			{
				JavaUnit javaunit= (JavaUnit) iter.next();
				javaunit.parse();
				panel.addClasses(javaunit.getClasses(), null);
			}
			catch (Exception e)
			{
				ErrorSystem.error("" + e);
			}
		}

		for (Iterator iter= widgethints.iterator(); iter.hasNext();)
		{
			ShapeHint wh= (ShapeHint) iter.next();
			panel.setShapeHint(wh);
		}

		for (Iterator iter= linkhints.iterator(); iter.hasNext();)
		{
			LinkHint lh= (LinkHint) iter.next();
			if(lh.type.equals("association"))
				panel.setLinkHint(lh.from, lh.to, Link.TYPE_ASSOCIATION);
			else if(lh.type.equals("generalization"))
				panel.setLinkHint(lh.from, lh.to, Link.TYPE_GENERALIZATION);	
		}

		for (Iterator iter= comments.iterator(); iter.hasNext();)
		{
			Comment comm= (Comment) iter.next();
			panel.setCommentHint(comm.comment, comm.name, comm.to, comm.x, comm.y, comm.lines);
		}
		panel.displayOn();
	}

	/** Returns all source paths of the coffeetronix project. */
	public Collection getProjectSources()
	{
		return srcdirs;
	}

	/** 
	 * Adds java files in a given directory, creates a java unit
		to commit the changes, all has to be cleared
		and consructPanel() activated.
	*/
	public void addJavaFiles(String fname)
	{

		try
		{
			if (srcdirs.contains(fname))
				return;

			srcdirs.add(fname);

			File f0= new File(projecthome + java.io.File.separatorChar + fname);

			if (f0 == null)
				return;
			if (!f0.exists())
				return;
			if (!f0.isDirectory())
				return;

			String filenames[]= f0.list(new FilenameFilter()
			{
				public boolean accept(File f, String s)
				{
					return s.endsWith(".java");
				}
			});

			for (int n= 0; n < filenames.length; n++)
			{
				// search for duplicate files
				File f1= new File(projecthome + File.separator + fname + File.separator + filenames[n]);

				for (Iterator iter= javaunits.iterator(); iter.hasNext();)
				{
					JavaUnit unit= (JavaUnit) iter.next();
					if (f1.getCanonicalPath().equals(unit.getFile().getCanonicalPath()))
						return;
				}
				javaunits.add(new JavaUnit(f1));
			}
		}
		catch (IOException e)
		{
		}
	}

	/** Saves the project to a file associated to the project */
	public boolean save()
	{
		return saveTo(file);
	}

	/** Saves the project to a file */
	public boolean saveTo(File file)
	{
		this.file= file;
		try
		{
			FileWriter fileout= new FileWriter(file);
			fileout.write("<!DOCTYPE CTXPROJECT>\n");
			fileout.write("<ctxproject>\n");
			fileout.write("<projecthome>" + projecthome + "</projecthome>\n");

			for (Iterator iter= srcdirs.iterator(); iter.hasNext();)
			{
				fileout.write("\t<srcdir>" + iter.next());
				fileout.write("</srcdir>\n");
			}

			fileout.write("\n");
			fileout.write("\t<config>\n");
			fileout.write("\t</config>\n");
			fileout.write("\n");

			panel.writeHints(fileout);

			fileout.write("</ctxproject>\n");

			fileout.flush();
			fileout.close();

		}
		catch (Exception e)
		{
			ErrorSystem.error("could not open fileoutput: " + file);
			return false;
		}

		ErrorSystem.info("saved project file: " + file);
		return true;
	}

	/**	Loads the project from a file. */
	public boolean load(File file)
	{
		this.file= file;

		return reload();
	}

	/**	Reloads the project. */
	public boolean reload()
	{
		init();

		Document doc;

		try
		{

			DocumentBuilderFactory docBuilderFactory= DocumentBuilderFactory.newInstance();
			DocumentBuilder docBuilder= docBuilderFactory.newDocumentBuilder();

			doc= docBuilder.parse(file);

			// normalize text representation
			doc.getDocumentElement().normalize();

			loadFromDoc(doc);

		}
		catch (SAXParseException err)
		{
			ErrorSystem.error("in parsing file " + file + ", line " + err.getLineNumber() + ", uri " + err.getSystemId());

			ErrorSystem.error("   " + err.getMessage());

		}
		catch (SAXException e)
		{
			Exception x= e.getException();

			((x == null) ? e : x).printStackTrace();

		}

		catch (Throwable t)
		{
			t.printStackTrace();
		}

		ErrorSystem.info("parsed project file: " + file);
		return true;
	}

	/** Processes DOM document to load project object. */
	private void loadFromDoc(Document doc)
	{
		Element root= doc.getDocumentElement();
		NodeList childs= root.getChildNodes();

		for (int n= 0; n < childs.getLength(); n++)
		{
			if (childs.item(n).getNodeType() == org.w3c.dom.Node.ELEMENT_NODE)
				processElement((Element) childs.item(n));
			// all other nodes are whitespace textnodes!
		}
	}

	/** Processes the xml-nodes during loading. */
	private void processElement(Element element)
	{
		if (element.getTagName().equals("projecthome"))
		{
			if (element.getChildNodes().getLength() == 1)
			{
				setProjectHome(element.getChildNodes().item(0).getNodeValue());
			}
			return;
		}

		if (element.getTagName().equals("srcdir"))
		{
			if (element.getChildNodes().getLength() == 1)
			{

				String fname= element.getChildNodes().item(0).getNodeValue();
				addJavaFiles(fname);
			}
			return;
		}

		if (element.getTagName().equals("shapehint"))
		{
			widgethints.add(
				new ShapeHint(
					element.getAttribute("name"),
					element.getAttribute("x"),
					element.getAttribute("y"),
					element.getAttribute("expanded"),
					element.getAttribute("color")));
			return;
		}

		if (element.getTagName().equals("linkhint"))
		{
			linkhints.add(
				new LinkHint(element.getAttribute("from"), element.getAttribute("to"), element.getAttribute("type")));
			return;
		}

		if (element.getTagName().equals("comment"))
		{
			StringBuffer str= new StringBuffer();

			int lines= 0;
			for (int n= 0; n < element.getChildNodes().getLength(); n++)
			{
				String s= element.getChildNodes().item(n).getNodeValue();
				if (s != null)
				{
					str.append(s + "\n");
					lines++;
				}
			}
			comments.add(
				new Comment(
					str.toString(),
					element.getAttribute("name"),
					element.getAttribute("to"),
					element.getAttribute("x"),
					element.getAttribute("y"),
					lines));
			return;
		}
	}

	public void setFileName(String name)
	{
		file=new File(name);
	}

	public void setProjectHome(String projecthome)
	{
		this.projecthome= projecthome;
	}

	public String getProjectHome()
	{
		return projecthome;
	}

	public void setSourceDirs(Collection c)
	{
		srcdirs= new LinkedList(c);
	}

	public void addSourceDir(String dir)
	{
		if (!srcdirs.contains(dir))
			srcdirs.add(dir);
	}

	public void removeSourceDir(String dir)
	{
		srcdirs.remove(dir);
	}
}
